%{

/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#include <unistd.h>


extern int yydebug ;

#define YYDEBUG 1

#include <string.h>
#include <errno.h>
#include "basename.hxx"
#include "yoo_root.hxx"
#include "dictionary.hxx"

extern FILE* yyin ;
extern FILE* yyout ;

#define SIZEOF_MESS 256 

static int SchErrorFlag ;
static char SchError[SIZEOF_MESS] ;
static char currclass[256] ;
static ListX<ADM_Class>* newclasses ;

int bracket_count = 0 ;
int linecount ;

char* infiles[512] ;
char* infile ;
char* outfile = NULL ;

extern "C" {
  int yywrap() ;
  int yyparse() ;
}

#if defined(__alpha__) || (defined(__sun__) && !defined(__svr4__))
int yylex() ;
#endif

#if defined(__alpha__) || (defined(__sun__) && !defined(__svr4__))
int yyoutput(int c) ; 
#else
extern "C" { int yyoutput(int c) ; }
#endif


class SchHead {
  private :

  char* _name ;
  char* _super ;
  
  public :

  SchHead(char* n, char* s = NULL) : _name(::strdup(n)) {
    if (s != NULL) {
      _super = ::strdup(s) ;
    } else {
      _super = NULL ;
    }
  }
  char* name() { return _name ; }
  char* super() { return _super ; }
} ;


extern "C" {
  void yoodaClassHeader() ;
}

typedef Iterator<ADM_SchFile> __Iter_ADM_SchFile ;

%}

%union {
  char                ident[256] ;
  int                 value ;
  SchHead*            classhead ;
} ;


%token <ident> IDENTIFIER LISTX
%token ACCESS SCHCLASS CBRACKET OBRACKET VIRTUAL SCHCLASSDECL PARAM EMPTYPARAM

%type  <classhead> schclass_head
%type  <ident> base_class
%type  <value> member local_member_list schclass

%start all

%%

all :
        schschema 

	
schschema :
        /* empty */
        | SCHCLASSDECL schschema
        | schclass CBRACKET
          {
	    if ($1 == TRUE) {
	      fprintf(yyout, ";\n\n#ifdef __YOODA_MAIN\n") ;
	      fprintf(yyout, "inline %s::%s() {}\n",currclass, currclass) ;
	      fprintf(yyout, "#endif\n") ;
	    }
          } schschema


schclass : 
	schclass_head local_member_list
          {
    	    ADM_Class* cl = YoodaDictionary->dictClass.getFirst($1->name()) ;
            if (cl == NULL) {
	      cl = new ADM_Class($1->name(),$1->super(),
					    YoodaDictionary->defClu) ;
	    }
            cl->yoodaGenStub(yyout) ;
            (*newclasses) += cl ;
	    if ($2 == FALSE) {
	      fprintf(yyout, "\n  %s() {}\n", $1->name()) ;
	    }
	    yyoutput('}') ;
	    $$ = $2 ;
	  } ;


schclass_head : 
	SCHCLASS IDENTIFIER {yyoutput(':'); yoodaClassHeader();} OBRACKET
          { 
            yyoutput('{') ;
	    $$ = new SchHead($2) ;
            strcpy(currclass, $2) ;
          }
	| SCHCLASS IDENTIFIER ':' base_class OBRACKET
          { 
            yyoutput('{') ;
            $$ = new SchHead($2, $4) ;
            strcpy(currclass, $2) ;
          } ;  


base_class : 
        inherit_mode IDENTIFIER
          {
	    strcpy($$, $2) ;
	  }


inherit_mode :
        /* empty */
        | ACCESS 
        | ACCESS VIRTUAL 
        | VIRTUAL ACCESS 
        | VIRTUAL ;

	
local_member_list :
        /* empty */
        {
	  $$ = FALSE ;
        }
	| member local_member_list
        {
  	  $$ = $1 | $2 ;
	}

	
member :
        IDENTIFIER EMPTYPARAM ';'
        {
	  if (strcmp($1, currclass) == 0) {
	    $$ = TRUE ;
	  } else {
    	    $$ = FALSE ;
	  }
	}
        | LISTX '<' IDENTIFIER '>' IDENTIFIER ';'
          {
	    $$ = FALSE ;
	  }
        | error ':'
          {
            SchErrorFlag = 0 ;
	    yyerrok ;
	    $$ = FALSE ;
	  } 
        | error OBRACKET
          {
            SchErrorFlag = 0 ;
	    yyerrok ;
	    $$ = FALSE ;
	  } 
        | error ';'
          {
            SchErrorFlag = 0 ;
	    yyerrok ;
	    $$ = FALSE ;
	  } ;


%%

void yyerror(char* s) 
{
  if (SchErrorFlag) {    
    printf("%s", SchError) ;
    yoo_exit(1) ;
  } else {
    sprintf(SchError,"%s:%d: %s\n", infile, linecount+1,s) ;
    SchErrorFlag = 1 ;
    return 1 ;
  }
}


void yoodaGenRefIncl(char* tmp)
{
  yyout = fopen(tmp, "w") ;
  Iterator<ADM_Class> iter(*newclasses) ;
  for (; iter.more(); iter.next()) {
    Ref<ADM_Class> cl = iter.get() ;
    fprintf(yyout, "#include \"ref_%s.hxx\"\n", (char*)cl->name) ;
  }
  fprintf(yyout, "\n\n") ;
  fclose(yyout) ;
}

extern "C" {

void yoodaClassHeader() 
{
  fprintf(yyout, " public ObjectX ") ;
}


void option_error(char* name) 
{
  printf("usage: %s -b <basename> <inputfile>\n", name);
  exit(1) ;
}


void yoodaGenBaseIncl()
{
  char f[256] ;
  sprintf(f, "%s/%s.hxx", YOO_SCH_DIR, YOODA_BASE) ;
  FILE* fincl = fopen(f, "w") ;
  Iterator<ADM_SchFile> iter(YoodaDictionary->dictSch) ;
  for (; iter.more(); iter.next()) {
    fprintf(fincl, "#include \"%s\"\n", (char*)iter.get()->name) ;
  }
  fclose(fincl) ;
}


void yoodaGenBaseSrc()
{
  char command[512];
  sprintf(command,"cp ${YOODA_HOME}/include/tmpl_pragma %s/%s.cc",
	  YOO_SCH_DIR,YOODA_BASE);
  system(command) ;

  char f[256] ;
  sprintf(f, "%s/%s.cc", YOO_SCH_DIR, YOODA_BASE) ;
  FILE* fsrc = fopen(f, "a") ;
  fprintf(fsrc, "#include \"%s.hxx\"\n", YOODA_BASE) ;
  fclose(fsrc) ;

  sprintf(command,"cat ${YOODA_HOME}/include/tmpl >> %s/%s.cc",
	  YOO_SCH_DIR,YOODA_BASE);
  system(command) ;
  
  fsrc = fopen(f, "a") ;
  fprintf(fsrc, "\n\n") ;
  Iterator<ADM_Class> iter(YoodaDictionary->dictClass) ;
  fprintf(fsrc, "  ADM_Class* cl ;\n") ;
  for (; iter.more(); iter.next()) {
    fprintf(fsrc,"  cl = YoodaDictionary->dictClass.getFirst(\"%s\");\n",
	    (char*)iter.get()->name);
    fprintf(fsrc, "  %s::yoo_ClId = cl->clId ;\n", (char*)iter.get()->name) ;
    fprintf(fsrc,"  if (cl == NULL) { \n") ;
    fprintf(fsrc,"    cerr << \"%s\" << \" : unknown class\" << endl;\n  }\n",
	    (char*)iter.get()->name);
    fprintf(fsrc,"  %s::yoo_Cluster=cl->myClu->clu;\n",
	    (char*)iter.get()->name) ;
    fprintf(fsrc, "  YOO_TabRefresh[cl->clId] = %s::yoo_refresh ;\n\n",
	    (char*)iter.get()->name) ;
  }
  fprintf(fsrc, "}\n\n") ;
  iter.reset() ;
  for (; iter.more(); iter.next()) {
    fprintf(fsrc, "Ref<Cluster> %s::yoo_Cluster ;\n", (char*)iter.get()->name);
    fprintf(fsrc, "int %s::yoo_ClId ;\n", (char*)iter.get()->name);
    fprintf(fsrc, "template class Ref<%s> ;\n",
	    (char*)iter.get()->name) ;
  }
    
  fprintf(fsrc, "\n\n#include \"%s_tmpl.hxx\"\n\n", YOODA_BASE) ;

  fprintf(fsrc, "void (*YOO_TabRefresh[%d]) (void*) ;\n",
	  YoodaDictionary->lastClass) ;
  fprintf(fsrc, "int YOO_LastClassId = %d;\n", YoodaDictionary->lastClass-1) ;
  fclose(fsrc) ;
}


int main(int argc, char** argv) {

  char* base = NULL ;
  int vers = 0 ;
  int cpt = 0 ;
  
  for (int i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-b") == 0) {
      base = argv[i+1] ;
      i++ ;
    } else if (strcmp(argv[i], "-v") == 0) {
      vers = 1 ;
    } else {
      infiles[cpt++] = argv[i] ;
    } 
  }

  if ( (base == NULL) || (cpt < 1) ) option_error(argv[0]) ;
  
  if (vers) {
    printf("YOODA version 1.2 <%s>, Copyright (C) 1994 Eric Abecassis
YOODA comes with ABSOLUTELY NO WARRANTY; for details type `COPYING'
in YOODA_HOME directory. This is free software, and you are welcome to 
redistribute it under certain conditions; for details read `COPYING'
or 'COPYING.LIB' in YOODA_HOME directory\n", argv[0]) ;
  }

  for (i = 0; i < cpt; i++) {
  
    yoo_connect(base, MONOUSER) ;

    newclasses = ::new ListX<ADM_Class> ;
    
    char* outfile = NULL ;
    outfile = new char[256] ;
    sprintf(outfile, "%s/%s.hxx", YOO_SCH_DIR, yoo_basename(infiles[i])) ;
    yyin = fopen(infiles[i], "r") ;
    infile = infiles[i] ;
    yyout = fopen(outfile, "w") ;

    if ( (!yyin) || (!yyout) ) {
      perror("") ;
      exit(1) ;
    }
    
    linecount = 0 ;
    SchErrorFlag = 0 ;
    yyparse() ;
    yyerror("") ;
    fclose(yyout) ;

    char command[512] ;
    char tmp[512] ;
    sprintf(tmp, "/usr/tmp/yooda_compiler_%s", YOODA_BASE) ;  
    yoodaGenRefIncl(tmp) ;
    sprintf(command,"cat %s >> %s; cp %s %s", outfile, tmp, tmp, outfile) ;
    system(command) ;
    unlink(tmp) ;
    
    sprintf(outfile, "%s.hxx", yoo_basename(infiles[i])) ;
    ADM_SchFile* sch = new ADM_SchFile(outfile) ;
    YoodaDictionary->dictSch.insertUniq(sch) ;

    yoodaGenBaseIncl() ;
    yoodaGenBaseSrc() ;

    yoo_commit() ;
    yoo_disconnect() ;
  }
}
	 
}
